/*
 * Copyright (C) 2014 Stuart Howarth <showarth@marxoft.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 3, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "onefichier.h"
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QRegExp>
#include <QTimer>
#ifdef DOWNLOAD_TEST
#include <QCoreApplication>
#include <QDebug>
#endif

OneFichier::OneFichier(QObject *parent) :
    ServicePlugin(parent),
    m_waitTimer(new QTimer(this)),
    m_waitTime(0),
    m_connections(1)
{
    this->connect(m_waitTimer, SIGNAL(timeout()), this, SLOT(updateWaitTime()));
#ifdef DOWNLOAD_TEST
    m_downloadReply = 0;
    this->setNetworkAccessManager(new QNetworkAccessManager(this));
    this->connect(this, SIGNAL(downloadRequestReady(QNetworkRequest)), this, SLOT(onDownloadRequestReady(QNetworkRequest)));
    this->connect(this, SIGNAL(error(ServicePlugin::ErrorType)), this, SLOT(onDownloadError(ServicePlugin::ErrorType)));
#endif
}

QRegExp OneFichier::urlPattern() const {
    return QRegExp("http(s|)://(\\w+.1fichier.com|1fichier.com/\\?\\w+)", Qt::CaseInsensitive);
}

bool OneFichier::urlSupported(const QUrl &url) const {
    return this->urlPattern().indexIn(url.toString()) == 0;
}

void OneFichier::login(const QString &username, const QString &password) {
    QString data = QString("mail=%1&pass=%2").arg(username).arg(password);
    QUrl url("https://1fichier.com/login.pl");
    QNetworkRequest request(url);
    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
    QNetworkReply *reply = this->networkAccessManager()->post(request, data.toUtf8());
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkLogin()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void OneFichier::checkLogin() {
    QNetworkReply *reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(NetworkError);
        return;
    }

    int statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

    switch (statusCode) {
    case 302:
    case 200:
    case 201:
        m_connections = 0;
        emit loggedIn(true);
        break;
    default:
        m_connections = 1;
        emit loggedIn(false);
        break;
    }

    reply->deleteLater();
}

void OneFichier::checkUrl(const QUrl &url) {
    QNetworkRequest request(url);
    request.setRawHeader("Accept-Language", "en-GB,en-US;q=0.8,en;q=0.6");
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkUrlIsValid()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void OneFichier::checkUrlIsValid() {
    QNetworkReply *reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit urlChecked(false);
        return;
    }

    QString redirect = reply->attribute(QNetworkRequest::RedirectionTargetAttribute).toString();
    QRegExp re("http(s|)://\\w-\\d.1fichier.com/[^'\"]+");

    if ((!redirect.isEmpty()) && (re.indexIn(redirect) == -1)) {
        this->checkUrl(QUrl(redirect));
    }
    else {
        QString response(reply->readAll());
        QString fileName = response.section("FileName :</td>", 1, 1)
                                   .section("class=\"normal\">", 1, 1)
                                   .section('<', 0, 0);

        if (fileName.isEmpty()) {
            emit urlChecked(false);
        }
        else {
            emit urlChecked(true, reply->request().url(), this->serviceName(), fileName);
        }
    }

    reply->deleteLater();
}

void OneFichier::getDownloadRequest(const QUrl &url) {
    emit statusChanged(Connecting);
    QNetworkRequest request(url);
    request.setRawHeader("Accept-Language", "en-GB,en-US;q=0.8,en;q=0.6");
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(onWebPageDownloaded()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void OneFichier::onWebPageDownloaded() {
    QNetworkReply *reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(NetworkError);
        return;
    }

    QRegExp re("http(s|)://\\w-\\d.1fichier.com/[^'\"]+");
    QString redirect = reply->attribute(QNetworkRequest::RedirectionTargetAttribute).toString();

    if (re.indexIn(redirect) == 0) {
        QNetworkRequest request;
        request.setUrl(QUrl(re.cap()));
        emit downloadRequestReady(request);
    }
    else if (!redirect.isEmpty()) {
        this->getDownloadRequest(QUrl(redirect));
    }
    else {
        QString response(reply->readAll());

        if (re.indexIn(response) >= 0) {
            QNetworkRequest request;
            request.setUrl(QUrl(re.cap()));
            emit downloadRequestReady(request);
        }
        else {
            QRegExp re("must wait \\d+ minutes");
        
            if (re.indexIn(response) >= 0) {
                int mins = re.cap().section("must wait ", 1, 1).section(' ', 0, 0).toInt();
                
                if (mins > 0) {
                    this->startWait(mins * 61000);
                    this->connect(this, SIGNAL(waitFinished()), this, SLOT(onWaitFinished()));
                }
                else {
                    emit error(UnknownError);
                }
            }
            else {
                this->getDownloadLink(reply->request().url());
            }
        }
    }

    reply->deleteLater();
}

void OneFichier::getDownloadLink(const QUrl &url) {
    QNetworkRequest request(url);
    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
    //request.setRawHeader("Referer", url.toString().toUtf8());
    //request.setRawHeader("Accept-Language", "en-GB,en-US;q=0.8,en;q=0.6");
    request.setRawHeader("User-Agent", "Wget/1.13.4 (linux-gnu)");
    request.setRawHeader("Connection", "Keep-Alive");
    QNetworkReply *reply = this->networkAccessManager()->post(request, QByteArray());
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkDownloadLink()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void OneFichier::checkDownloadLink() {
    QNetworkReply *reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(NetworkError);
        return;
    }

    QUrl redirect = reply->attribute(QNetworkRequest::RedirectionTargetAttribute).toUrl();

    if (!redirect.isEmpty()) {
        emit downloadRequestReady(QNetworkRequest(redirect));
    }
    else {
        QString response(reply->readAll());
        QRegExp re("must wait \\d+ minutes");
        
        if (re.indexIn(response) >= 0) {
            int mins = re.cap().section("must wait ", 1, 1).section(' ', 0, 0).toInt();
            
            if (mins > 0) {
                this->startWait(mins * 61000);
                this->connect(this, SIGNAL(waitFinished()), this, SLOT(onWaitFinished()));
            }
            else {
                emit error(UnknownError);
            }
        }
        else {
            emit error(UnknownError);
        }
    }

    reply->deleteLater();
}

void OneFichier::startWait(int msecs) {
#ifdef DOWNLOAD_TEST
    qDebug() << "Need to wait" << msecs << "before download";
    QCoreApplication::instance()->quit();
#else
    if (msecs > 60000) {
        emit statusChanged(LongWait);
    }
    else {
        emit statusChanged(ShortWait);
    }

    emit waiting(msecs);
    m_waitTime = msecs;
    m_waitTimer->start(1000);
#endif
}

void OneFichier::updateWaitTime() {
    m_waitTime -= m_waitTimer->interval();
    emit waiting(m_waitTime);

    if (m_waitTime <= 0) {
        m_waitTimer->stop();
        emit waitFinished();
    }
}

void OneFichier::onWaitFinished() {
    emit statusChanged(Ready);
    this->disconnect(this, SIGNAL(waitFinished()), this, SLOT(onWaitFinished()));
}

bool OneFichier::cancelCurrentOperation() {
    m_waitTimer->stop();
    this->disconnect(this, SIGNAL(waitFinished()), this, 0);
    emit currentOperationCancelled();

    return true;
}

#ifdef DOWNLOAD_TEST
void OneFichier::downloadTest(const QUrl &url, const QString &fileName) {
    m_downloadFile.setFileName(fileName);
    this->getDownloadLink(url);
}
    
void OneFichier::onDownloadRequestReady(const QNetworkRequest &request) {
    if (m_downloadReply) {
        delete m_downloadReply;
    }
    
    if (!m_downloadFile.open(QIODevice::WriteOnly)) {
        qDebug() << "Download error:" << m_downloadFile.errorString();
        QCoreApplication::instance()->quit();
    }
    
    m_downloadReply = this->networkAccessManager()->get(request);
    this->connect(m_downloadReply, SIGNAL(downloadProgress(qint64,qint64)), this, SLOT(onDownloadProgressChanged(qint64,qint64)));
    this->connect(m_downloadReply, SIGNAL(readyRead()), this, SLOT(onDownloadReadyRead()));
    this->connect(m_downloadReply, SIGNAL(finished()), this, SLOT(onDownloadFinished()));
    this->connect(m_downloadReply, SIGNAL(sslErrors(QList<QSslError>)), this, SLOT(onDownloadSSLErrors(QList<QSslError>)));
}

void OneFichier::onDownloadProgressChanged(qint64 received, qint64 total) {
    qDebug() << "Download progress:" << received << "/" << total;
}

void OneFichier::onDownloadReadyRead() {
    m_downloadFile.write(m_downloadReply->readAll());
}

void OneFichier::onDownloadFinished() {
    m_downloadFile.close();
    
    switch (m_downloadReply->error()) {
    case QNetworkReply::NoError:
        qDebug() << "Download completed successfully at" << m_downloadFile.fileName();
        break;
    default:
        qDebug() << "Download error:" << m_downloadReply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt() << m_downloadReply->errorString();
        break;
    }
    
    m_downloadReply->deleteLater();
    m_downloadReply = 0;
    //QCoreApplication::instance()->quit();
}

void OneFichier::onDownloadError(ServicePlugin::ErrorType e) {
    qDebug() << "Download error:" << e;
    QCoreApplication::instance()->quit();
}

void OneFichier::onDownloadSSLErrors(const QList<QSslError> &errors) {
    qDebug() << "SSL errors:" << errors;
    m_downloadReply->ignoreSslErrors(errors);
}
#endif

Q_EXPORT_PLUGIN2(onefichier, OneFichier)
